#include <stdio.h>
#include <MenuItem.h>
#include <Alert.h>
#include "ObjectInspectorView.h"
#include "ObjectInspectorItem.h"
#include "ColumnListView.h"
#include "BerryUtils.h"
#include "Locale.h"


// TItemEditor -----------------------------------------------------------------------------

TItemEditor::TItemEditor(ColumnListView *owner,
	BRect frame,
	const char *name,
	const char *label,
	const char *text,
	const selectable_values *selectableValues,
	uint32 resizingMode,
	uint32 flags)
	:	BTextControl(frame, name, label, text, NULL, resizingMode, flags)
{
	fOwner = owner;
	fSelectableValues = selectableValues;
	fPopUpMenu = NULL;
	if (fSelectableValues != NULL) {
		this->SetDivider(10);
		fPopUpMenu = new BPopUpMenu("PopUpMenu");
		fPopUpMenu->SetRadioMode(false);
		BMenuItem *item;
		for(uint32 i=0; fSelectableValues->SelectableStrings[i] != NULL; i++) {
			item = new BMenuItem(fSelectableValues->SelectableStrings[i], NULL);
			fPopUpMenu->AddItem(item);
		}
	} else {
		this->SetDivider(0);
	}
}

TItemEditor::~TItemEditor()
{
	if (fPopUpMenu != NULL) delete fPopUpMenu;
}

void TItemEditor::AttachedToWindow()
{
	fOriginalText.SetTo(Text());
}

bool TItemEditor::IsModified()
{
	return fOriginalText.Compare(Text()) != 0;
}

void TItemEditor::ResetText()
{
	fOriginalText.SetTo(Text());
}

void TItemEditor::UpdateMenuItemMark()
{
	if (fSelectableValues == NULL) return;
	
	if ((fSelectableValues->FlagMode) &&
	    (fSelectableValues->Separator != NULL)) {
	    TStringEx sep(fSelectableValues->Separator);
	    sep.Trim();
		TStrTok strTok(this->Text(), sep.String(), 0);
		for(int i=0; i < strTok.CountTokens(); i++) {
			TStringEx str(strTok.Token(i));
			str.Trim();
			if (str.Length()<=0) continue;
			BMenuItem *item = fPopUpMenu->FindItem(str.String());
			if (item != NULL) item->SetMarked(true);
		}
	}
}

void TItemEditor::MouseDown(BPoint point)
{
	if (fPopUpMenu == NULL) return;
	
	int32 buttons;
	this->Window()->CurrentMessage()->FindInt32("buttons", &buttons);
	if ((buttons & B_SECONDARY_MOUSE_BUTTON)!= 0) {
		BMenuItem *selected;
		ConvertToScreen(&point);
		selected = fPopUpMenu->Go(point, false, true, BRect(point.x, point.y, point.x, point.y));
		if (selected != NULL) {
			if (fSelectableValues->FlagMode) {
				selected->SetMarked(!selected->IsMarked());
				BString str;
				const char *sep = fSelectableValues->Separator;
				for(int32 i=0; i < fPopUpMenu->CountItems(); i++) {
					BMenuItem *item = fPopUpMenu->ItemAt(i);
					if (item->IsMarked()) {
						str << (str.Length() > 0 ? sep : "" ) << item->Label();
					}
				}
				this->SetText(str.String());
			} else {
				this->SetText(selected->Label());
			}
			BMessenger(fOwner).SendMessage(ENTER_MSG);
		}
	}
}


// TItemEditMessageFilter ---------------------------------------------------------------

TItemEditMessageFilter::TItemEditMessageFilter(ColumnListView *owner)
	:	BMessageFilter(B_ANY_DELIVERY, B_ANY_SOURCE)
{
	fOwner = owner;
}

TItemEditMessageFilter::~TItemEditMessageFilter()
{
}

filter_result TItemEditMessageFilter::Filter(BMessage *message, BHandler **handler)
{
	switch(message->what) {
		case B_KEY_DOWN:
//			message->PrintToStream();
			int8 kvalue;
			if (message->FindInt8("byte", 0, &kvalue)==B_OK) {
				switch(kvalue) {
					case B_ENTER:
						BMessenger(fOwner).SendMessage(ENTER_MSG);
						break;
					case B_DOWN_ARROW:
						BMessenger(fOwner).SendMessage(GONEXT_MSG);
						break;
					case B_UP_ARROW:
						BMessenger(fOwner).SendMessage(GOPREV_MSG);
						break;
					case 16:	// F1 key
						int32 key;
						if (message->FindInt32("key", &key) == B_OK) {
							if (key == 2) BMessenger(fOwner).SendMessage(GB_SHOW_HELP);
						}
						break;
				}
			}
			break;
	}
	return B_DISPATCH_MESSAGE;
}


// TObjectInspectorItem ------------------------------------------------------------------

TObjectInspectorItem::TObjectInspectorItem(
			ColumnListView *owner,
			const TPropertyItem *PropertyItem,
			int32 level,
			BList *objectList,
			bool superitem,
			TBerryComponentLibrary *bcl
		)
	:	CLVEasyItem(level, superitem, false, 20)
{
	fPropertyItem = *PropertyItem;
	fBCL = bcl;
	fEditor = NULL;
	fOwner = owner;
	fOpenPropEditorBtn = NULL;
	fObjectList.AddList(objectList);
	SetColumnContent(COLUMN_NAME, PropertyItem->Name);
	SetColumnContent(COLUMN_VALUE, "");
}

TObjectInspectorItem::~TObjectInspectorItem()
{
	Deselected();
}

void TObjectInspectorItem::Selected()
{
	if (IsSuperItem()) {
		fOwner->MakeFocus();
		return;
	}
	
	BRect rect = this->ItemColumnFrame(COLUMN_VALUE, fOwner);
	rect.top += 1;
	rect.right -= (fPropertyItem.TypeInfo != NULL) ? 16 : 0;
	fEditor = new TItemEditor(fOwner, rect, "edit", ">", "", fPropertyItem.SelectableValues);
	fEditor->TextView()->MakeEditable(fPropertyItem.TypeCode != B_RAW_TYPE);
	fEditor->TextView()->AddFilter((BMessageFilter *)new TItemEditMessageFilter(fOwner));
	fEditor->SetText(GetColumnContentText(COLUMN_VALUE));
	fOwner->AddChild(fEditor);
	fEditor->MakeFocus();
	fEditor->TextView()->SelectAll();
	fEditor->UpdateMenuItemMark();
	
	if (fPropertyItem.TypeInfo != NULL) {
		BRect frame = rect;
		frame.left = rect.right + 2;
		frame.right = frame.left + 16;
		BMessage *msg = new BMessage(OPEN_PROP_EDITOR);
		msg->AddPointer("prop_editor", (void *)fPropertyItem.TypeInfo);
		fOpenPropEditorBtn = new BButton(frame, "OpenPropEditorBtn", "...", msg);
		fOpenPropEditorBtn->SetTarget(fOwner);
		fOpenPropEditorBtn->ResizeBy(0, -6);
		fOwner->AddChild(fOpenPropEditorBtn);
	}
}

void TObjectInspectorItem::Deselected()
{
	if (fEditor!=NULL) {
		fOwner->RemoveChild(fEditor);
		delete fEditor;
		fEditor = NULL;
		if (fPropertyItem.TypeInfo != NULL) {
			fOwner->RemoveChild(fOpenPropEditorBtn);
			delete fOpenPropEditorBtn;
			fOpenPropEditorBtn = NULL;
		}
		fOwner->Invalidate();
	}
}

const char *TObjectInspectorItem::EditorText()
{
	if (fEditor==NULL) return NULL;
	return fEditor->Text();
}

void TObjectInspectorItem::UpdateContent()
{
	if (IsSuperItem()) return;
	
	BString value, lastValue;
	BArchivable *aObject;
	status_t status;
	TBerryComponentLibrary *bcl;
	
	for(int32 i=0; i<fObjectList.CountItems(); i++) {
		aObject = (BArchivable *)fObjectList.ItemAt(i);
		if (fBCL == NULL) {
			FindSupportBCL(aObject, &bcl);
		} else  {
			bcl = fBCL;
		}
		if (bcl!=NULL) {
			value.SetTo("");
			LockObjectLooper(aObject);
			status = bcl->GetValue(aObject, GetColumnContentText(COLUMN_NAME), &value);
			UnlockObjectLooper(aObject);
			if (status!=B_OK) {
				value.SetTo("");
				break;
			}
			if (lastValue.Length()!=0) if (lastValue.Compare(value)!=0) {
				value.SetTo("");
				break;
			}
			lastValue = value;
		} else {
			value.SetTo("");
		}
	}
	SetColumnContent(COLUMN_VALUE, value.String());
	if (fEditor!=NULL) {
		fEditor->SetText(value.String());
		fEditor->TextView()->SelectAll();
	}
	fOwner->InvalidateItem(fOwner->IndexOf(this));
}

void TObjectInspectorItem::SetValueFromItemEditor(bool noAlert)
{
	if (fEditor==NULL) return;
	if (!fEditor->IsModified()) return;
	this->SetValue(fEditor->Text(), noAlert);
}

void TObjectInspectorItem::SetValue(const char *value, bool noAlert)
{
	TBerryComponentLibrary *bcl;
	BArchivable *aObject;
	for(int32 i=0; i<fObjectList.CountItems(); i++) {
		aObject = (BArchivable *)fObjectList.ItemAt(i);
		if (aObject == NULL) continue;
		if (fBCL == NULL) {
			FindSupportBCL(aObject, &bcl);
		} else  {
			bcl = fBCL;
		}
		if (bcl != NULL) {
			BString _value;
			const char *propertyName = this->GetColumnContentText(COLUMN_NAME);
			LockObjectLooper(aObject);
			if (bcl->SetValue(aObject, propertyName, value)==B_OK) {
				if (bcl->GetValue(aObject, propertyName, &_value)==B_OK) {
					if (fEditor!=NULL) {
						fEditor->SetText(_value.String());
						fEditor->TextView()->SelectAll();
						fEditor->UpdateMenuItemMark();
						fEditor->ResetText();
					}
					this->SetColumnContent(COLUMN_VALUE, _value.String());
				}
			} else {
				fprintf(stderr, APP_NAME ": WARNING _updater_func() invalid value (%s).\n", this->EditorText());
				if (!noAlert) {
//					fOwner->Window()->CurrentMessage()->PrintToStream();
					if (fOwner->Window()->CurrentMessage()->what != 'bvsl') {
						BString alertText(GET_STRING("InvalidValue", "\"%s\" is an invalid value."));
						alertText.ReplaceFirst("%s", this->EditorText());
						(new BAlert("", alertText.String(), GET_STRING("OK", "OK")))->Go();
					}
				}
				if (bcl->GetValue(aObject, propertyName, &_value)==B_OK) {
					this->SetColumnContent(COLUMN_VALUE, _value.String());
					if (fEditor!=NULL) {
						fEditor->SetText(_value.String());
						fEditor->TextView()->SelectAll();
					}
				} else {
					if (fEditor!=NULL) {
						fEditor->SetText("");
					}
				}
			}
			UnlockObjectLooper(aObject);
		}
	}

}

BArchivable *TObjectInspectorItem::Object(int32 index) const
{
	return (BArchivable *)fObjectList.ItemAt(index);
}

