#include <stdio.h>
#include "ToolTip.h"


TToolTip tool_tip;


// TToolTipWindow

TToolTipWindow::TToolTipWindow()
	:	BWindow(BRect(-1,-1,-1,-1), "ToolTip",
				B_NO_BORDER_WINDOW_LOOK, B_NORMAL_WINDOW_FEEL, B_AVOID_FOCUS)
{
	fRunner = new BMessageRunner(BMessenger(this), new BMessage('incr'), 100000);
	fRunnerEnabled = false;
	
	BGView = new BView(Bounds(), "BGView", B_FOLLOW_ALL_SIDES, B_WILL_DRAW | B_FRAME_EVENTS);
	BGView->SetViewColor(0, 0, 0);
	AddChild(BGView);
	
	BRect rect(1, 1, BGView->Frame().right - 1, BGView->Frame().bottom - 1);
	TipText = new BTextView(rect, "TipText", rect,
							B_FOLLOW_ALL_SIDES, B_WILL_DRAW | B_FRAME_EVENTS);
	TipText->SetWordWrap(false);
	TipText->SetViewColor(0xf0, 0xf0, 0);
	BGView->AddChild(TipText);
	
	fPulseCount = 0;
	Target = NULL;
	fRunner = NULL;
	
	Minimize(true);
	Show();
}

TToolTipWindow::~TToolTipWindow()
{
	delete fRunner;
}

void TToolTipWindow::SetToolTip(BView *target, const char *tip)
{
	if (target->Window() == NULL) return;
	
	Target = target;
	target->Window()->Lock();
	BRect targetFrame(target->ConvertToScreen(target->Bounds()));
	target->Window()->Unlock();
	TipText->SetText(tip);
	
	// TipText の全テキストサイズからウィンドゥを移動・リサイズする
	float width = 0;
	for(int32 i=0; i < TipText->CountLines(); i++) {
		width = max_c(width, TipText->LineWidth(i));
	}
	width += 2;
	float height = TipText->TextHeight(0, TipText->CountLines()-1);
	MoveTo(targetFrame.left + (targetFrame.Width() / 2) - (width / 2), targetFrame.bottom + 5);
	ResizeTo(width, height);
	TipText->SetTextRect(BRect(0, 0, width - 1, height - 1));
}

void TToolTipWindow::MessageReceived(BMessage *msg)
{
	switch(msg->what){
		case 'sttp': {
			void *view;
			BString tip;
			if ((msg->FindPointer("target", &view) == B_OK) &&
			    (msg->FindString("tip", &tip) == B_OK)) {
				SetToolTip(reinterpret_cast<BView *>(view), tip.String());
				fRunnerEnabled = true;
			}
			break;
		}
		case 'puls': {
			BPoint loc;
			uint32 button;
			BGView->GetMouse(&loc, &button);
			ConvertToScreen(&loc);
			if (loc != tool_tip.CursorPos) fPulseCount = 0;
			fRunnerEnabled = true;
			break;
		}
		case 'incr':
			if (!fRunnerEnabled) break;
			ShowToolTip();
			break;
		default: BWindow::MessageReceived(msg);
	}
}

void TToolTipWindow::CloseToolTip()
{
	Minimize(true);
	fRunnerEnabled = false;
}

void TToolTipWindow::ShowToolTip()
{
	if (Target == NULL) {
		CloseToolTip();
		return;
	}
	BWindow *window = Target->Window();
	if (window == NULL) {
		CloseToolTip();
		return;
	}
	
	BPoint loc;
	uint32 button;
	BGView->GetMouse(&loc, &button);
	ConvertToScreen(&loc);
	fPulseCount++;
	if ((fPulseCount > 5) && (IsMinimized())) {
		BRect targetRect;
		if (window->Lock()) {
			targetRect = Target->Bounds();
			Target->ConvertToScreen(&targetRect);
			window->Unlock();
		} else {
			fprintf(stderr, "ToolTip: window->Lock() failed.\n");
			return;
		}
		if (targetRect.Contains(loc)) {
			Minimize(false);
		} else {
			CloseToolTip();
		}
	} else {
		if (fPulseCount > 10) {
			if (loc != tool_tip.CursorPos) {
				CloseToolTip();
			}
		}
	}
}


// TToolTipFilter

TToolTipFilter::TToolTipFilter(BView *target)
	:	BMessageFilter(B_ANY_DELIVERY, B_ANY_SOURCE)
{
	Target = target;
}

TToolTipFilter::~TToolTipFilter()
{
}

filter_result TToolTipFilter::Filter(BMessage *msg, BHandler **target)
{
	switch(msg->what) {
		case B_MOUSE_MOVED:
			if (msg->FindPoint("where", &tool_tip.CursorPos) == B_OK) {
				if (Target->Window()->Lock()) {
					Target->Window()->ConvertToScreen(&tool_tip.CursorPos);
					Target->Window()->Unlock();
					ShowToolTip();
				}
			}
			break;
	}
	return B_DISPATCH_MESSAGE;
}

void TToolTipFilter::SetToolTip(const char *tip)
{
	ToolTip.SetTo(tip);
}

void TToolTipFilter::ShowToolTip()
{
	if (tool_tip.LastToolTipTargetView != Target) {
		tool_tip.LastToolTipTargetView = Target;
		BMessage msg('sttp');
		msg.AddPointer("target", Target);
		msg.AddString("tip", ToolTip);
		BMessenger(tool_tip.ToolTipWindow).SendMessage(&msg);
	} else {
		BMessage msg('puls');
		BMessenger(tool_tip.ToolTipWindow).SendMessage(&msg);
	}
}

// TToolTip

TToolTip::TToolTip()
{
	LastToolTipTargetView = NULL;
	ToolTipWindow = NULL;
}

TToolTip::~TToolTip()
{
	if (ToolTipWindow != NULL) BMessenger(ToolTipWindow).SendMessage(B_QUIT_REQUESTED);
}

status_t TToolTip::SetToolTip(BView *target, const char *tip)
{
	if (ToolTipWindow == NULL) ToolTipWindow = new TToolTipWindow();
	
	// target から TToolTipFilter を探しだす
	BList *filterList = target->FilterList();
	TToolTipFilter *targetFilter = NULL;
	if (filterList != NULL) {
		for(int32 i=0; i < filterList->CountItems(); i++) {
			BMessageFilter *filter = (BMessageFilter *)filterList->ItemAt(i);
			if (dynamic_cast<TToolTipFilter *>(filter) != NULL) {
				targetFilter = (TToolTipFilter *)filter;
				break;
			}
		}
	}
	
	// もし TToolTipFilter が無ければ追加
	if (targetFilter == NULL) {
		targetFilter = new TToolTipFilter(target);
		target->AddFilter(targetFilter);
	}
	
	targetFilter->SetToolTip(tip);
	return B_OK;
}
